from arch_lint.graph import (
    FullPathModule,
    ImportGraph,
)
from typing import (
    Dict,
    FrozenSet,
    NoReturn,
    Union,
)


def unauthorized_imports(
    forbidden: FullPathModule,
    allowlist: FrozenSet[FullPathModule],
    graph: ImportGraph,
) -> FrozenSet[FullPathModule]:
    return graph.find_modules_that_directly_import(forbidden) - allowlist


def check_forbidden(
    forbidden_allow_list: Dict[FullPathModule, FrozenSet[FullPathModule]],
    graph: ImportGraph,
) -> Union[None, NoReturn]:
    for forbidden, allowlist in forbidden_allow_list.items():
        illegal = unauthorized_imports(forbidden, allowlist, graph)
        if illegal:
            _illegal = frozenset(i.module for i in illegal)
            raise Exception(
                f"Forbidden module `{forbidden.module}` imported by unauthorized modules i.e. {_illegal}"
            )
    return None



from __future__ import (
    annotations,
)

from dataclasses import (
    dataclass,
)
import grimp
from grimp.application.ports.graph import (
    AbstractImportGraph,
)
from importlib.util import (
    find_spec,
)
from typing import (
    cast,
    FrozenSet,
    NoReturn,
    Union,
)


@dataclass(frozen=True)
class _FullPathModule:
    pass


@dataclass(frozen=True)
class FullPathModule:
    _inner: _FullPathModule
    module: str

    @property
    def name(self) -> str:
        return self.module.split(".")[-1]

    def new_child(self, module: str) -> FullPathModule:
        joined = ".".join([self.module, module])
        return FullPathModule(_FullPathModule(), joined)

    @staticmethod
    def from_raw(raw: str) -> Union[FullPathModule, NoReturn]:
        spam_spec = find_spec(raw)
        if spam_spec is not None:
            return FullPathModule(_FullPathModule(), raw)
        raise ModuleNotFoundError(raw)


def _build_module(module: str) -> FullPathModule:
    return FullPathModule(_FullPathModule(), module)


@dataclass(frozen=True)  # type: ignore[misc]
class _ImportGraph:  # type: ignore[no-any-unimported]
    graph: AbstractImportGraph  # type: ignore[no-any-unimported]


@dataclass(frozen=True)
class ImportGraph:
    _inner: _ImportGraph
    root: FullPathModule

    @staticmethod
    def build_graph(
        root_module: str, external_packages: bool
    ) -> Union[ImportGraph, NoReturn]:
        graph = grimp.build_graph(root_module, include_external_packages=external_packages)  # type: ignore[misc]
        return ImportGraph(_ImportGraph(graph), _build_module(root_module))  # type: ignore[misc]

    def chain_exists(
        self,
        importer: FullPathModule,
        imported: FullPathModule,
        as_packages: bool,
    ) -> bool:
        return cast(
            bool,
            self._inner.graph.chain_exists(importer.module, imported.module, as_packages),  # type: ignore[misc]
        )

    def find_children(
        self, module: FullPathModule
    ) -> FrozenSet[FullPathModule]:
        items: FrozenSet[str] = frozenset(self._inner.graph.find_children(module.module))  # type: ignore[misc]
        return frozenset(_build_module(i) for i in items)

    def find_modules_that_directly_import(
        self, module: FullPathModule
    ) -> FrozenSet[FullPathModule]:
        items: FrozenSet[str] = frozenset(self._inner.graph.find_modules_that_directly_import(module.module))  # type: ignore[misc]
        return frozenset(_build_module(i) for i in items)



from dataclasses import (
    dataclass,
)
from typing import (
    Dict,
    FrozenSet,
    NoReturn,
    Optional,
    Tuple,
    Union,
)


@dataclass(frozen=True)
class _DAG:
    items: Dict[str, Tuple[FrozenSet[str], ...]]


@dataclass(frozen=True)
class DAG:
    _inner: _DAG

    def get(self, module: str) -> Optional[Tuple[FrozenSet[str], ...]]:
        return self._inner.items.get(module)

    def get_children(self, module: str) -> FrozenSet[str]:
        items = self.get(module)
        if items:
            item_set = set()
            for i in items:
                for x in i:
                    if x not in item_set:
                        item_set.add(x)
                    else:
                        raise Exception(
                            f"Dag must not contain duplicated child modules i.e. {module + '.' + x}"
                        )
            return frozenset(item_set)
        return frozenset()


def _assert_set(items: Tuple[str, ...]) -> FrozenSet[str]:
    if len(items) == len(frozenset(items)):
        return frozenset(items)
    raise ValueError("Expected a set but got duplicated values")


def _to_set(item: Union[Tuple[str, ...], str]) -> FrozenSet[str]:
    if isinstance(item, tuple):
        return _assert_set(item)
    return _assert_set((item,))


def new_dag(
    raw: Dict[str, Tuple[Union[Tuple[str, ...], str], ...]]
) -> Union[DAG, NoReturn]:
    _raw = {k: tuple(_to_set(i) for i in v) for k, v in raw.items()}
    return DAG(_DAG(_raw))



from .core import (
    DAG,
)
from arch_lint.graph import (
    FullPathModule,
    ImportGraph,
)
from typing import (
    FrozenSet,
    NoReturn,
    Optional,
    Tuple,
    Union,
)


def _chain_exist(
    graph: ImportGraph,
    importer: FrozenSet[FullPathModule],
    imported: FrozenSet[FullPathModule],
) -> Optional[Tuple[FullPathModule, FullPathModule]]:
    for s in importer:
        for t in imported:
            if graph.chain_exists(s, t, True):
                return (s, t)
    return None


def _independence(
    graph: ImportGraph, modules: FrozenSet[FullPathModule]
) -> Optional[Tuple[FullPathModule, FullPathModule]]:
    for m in modules:
        checks = modules - frozenset([m])
        for c in checks:
            if graph.chain_exists(m, c, True):
                return (m, c)
    return None


def _check_independence(
    graph: ImportGraph, modules: FrozenSet[FullPathModule]
) -> Union[None, NoReturn]:
    result = _independence(graph, modules)
    if result:
        raise Exception(
            f"Broken DAG same lvl modules should be independent {result[0].module} -> {result[1].module}"
        )
    return None


def _check_dag_over_modules(
    graph: ImportGraph,
    lower: FrozenSet[FullPathModule],
    upper: FrozenSet[FullPathModule],
) -> Union[None, NoReturn]:
    _chain = _chain_exist(graph, upper, lower)
    if _chain:
        raise Exception(
            f"Broken DAG with illegal import {_chain[0].module} -> {_chain[1].module}"
        )
    return None


def check_dag(
    dag: DAG, graph: ImportGraph, module: FullPathModule
) -> Union[None, NoReturn]:
    _children = graph.find_children(module)
    children = frozenset(s.name for s in _children)
    dag_modules = dag.get_children(module.module)
    missing = children - dag_modules
    if len(_children) > 1 and missing:
        raise Exception(
            f"Missing children modules of `{module.module}` at DAG i.e. {missing}"
        )
    dag_struct = dag.get(module.module)
    if dag_struct:
        for n, c in enumerate(dag_struct):
            current_modules = frozenset(module.new_child(x) for x in c)
            _check_independence(graph, current_modules)
            for upper_modules in dag_struct[n + 1 :]:
                _check_dag_over_modules(
                    graph,
                    current_modules,
                    frozenset(module.new_child(x) for x in upper_modules),
                )
    for child in _children:
        check_dag(dag, graph, child)

    return None



from arch_lint.dag.core import (
    DAG,
    new_dag,
)
from arch_lint.graph import (
    FullPathModule,
)
from typing import (
    Dict,
    FrozenSet,
    Tuple,
    Union,
)

_dag: Dict[str, Tuple[Union[Tuple[str, ...], str], ...]] = {
    "arch_lint": (
        ("dag", "forbidden", "private"),
        "graph",
    ),
    "arch_lint.dag": (
        "check",
        "core",
    ),
}


def project_dag() -> DAG:
    return new_dag(_dag)


def forbidden_allowlist() -> Dict[FullPathModule, FrozenSet[FullPathModule]]:
    _raw: Dict[str, FrozenSet[str]] = {
        "grimp": frozenset(["arch_lint.graph"]),
    }
    return {
        FullPathModule.from_raw(k): frozenset(
            FullPathModule.from_raw(i) for i in v
        )
        for k, v in _raw.items()
    }



from .arch import (
    forbidden_allowlist,
    project_dag,
)
from arch_lint.dag.check import (
    check_dag,
)
from arch_lint.forbidden import (
    check_forbidden,
)
from arch_lint.graph import (
    ImportGraph,
)
from arch_lint.private import (
    check_private,
)


def test_dag_creation() -> None:
    project_dag()


def test_dag() -> None:
    root = "arch_lint"
    graph = ImportGraph.build_graph(root, True)
    check_dag(project_dag(), graph, graph.root)


def test_forbidden_creation() -> None:
    forbidden_allowlist()


def test_forbidden() -> None:
    root = "arch_lint"
    graph = ImportGraph.build_graph(root, True)
    allowlist_map = forbidden_allowlist()
    check_forbidden(allowlist_map, graph)


def test_private() -> None:
    root = "arch_lint"
    graph = ImportGraph.build_graph(root, False)
    check_private(graph, graph.root)



from arch_lint.graph import (
    FullPathModule,
    ImportGraph,
)
from typing import (
    NoReturn,
    Union,
)


def _check_private(
    graph: ImportGraph, parent: FullPathModule, module: FullPathModule
) -> None:
    importers = graph.find_modules_that_directly_import(module)
    is_private = module.name.startswith("_")

    def _valid_importer(importer: FullPathModule) -> bool:
        return importer.module.startswith(parent.module)

    if is_private:
        for i in importers:
            if not _valid_importer(i):
                raise Exception(
                    f"Illegal import of private module {i.module} -> {module.module}"
                )
    for c in graph.find_children(module):
        _check_private(graph, module, c)


def check_private(
    graph: ImportGraph, module: FullPathModule
) -> Union[None, NoReturn]:
    for c in graph.find_children(module):
        _check_private(graph, module, c)
    return None
